/* ---------------------------------------------------------------- */
/* MUTEX_INIT():                                                    */
/*   Creates a Mutex lock.                                          */
/*                                                                  */
/* Return value:                                                    */
/*   Mutex's ID     -- Success                                      */
/*   mtu_ERROR      -- Failure                                      */
/* ---------------------------------------------------------------- */

MUTEX_t MUTEX_INIT(void)
{                                    
     struct MUTEX *NewMutex;

     NewMutex = (struct MUTEX *) malloc(sizeof(struct MUTEX));
     if (NewMutex == NULL) 
          return (MUTEX_t) mtu_ERROR;

     NewMutex->Owner = (THREAD_t)mtu_NOBODY;
     NewMutex->WaitingQ.Head = NewMutex->WaitingQ.Tail = NULL;
     NewMutex->Magic = mtu_VALID;       /* set a magic number       */  
     THREAD_Append(&SysTable, (void *)NewMutex);
     return NewMutex;
}

/* ---------------------------------------------------------------- */
/* MUTEX_DESTROY():                                                 */
/*   Destroys a Mutex lock.                                         */
/*                                                                  */
/* Argument:                                                        */
/*   MUTEX_t Id     -- the mutex ID to be destroyed                 */
/* Return value:                                                    */
/*   mtu_NORMAL     -- Success                                      */
/*   mtu_ERROR      -- Failure                                      */
/* ---------------------------------------------------------------- */

int MUTEX_DESTROY(MUTEX_t Id)
{
     if (SYSTEM_INITIALIZE == FALSE) {
       fprintf(stderr, "Error in MUTEX_DESTROY() ");
       fprintf(stderr, "MUTEX hasn't been initialized yet\n");
       exit(0);
     }

     if (Id->Magic != mtu_VALID) {       /* invalid mutex ID        */
          mtuMTP_errno = mtuMTP_INVALMUTEX;
          return mtu_ERROR;
     }
     THREAD_RemovefromQueue(&SysTable, Id);
     free(Id);
     return mtu_NORMAL;
}

/* ---------------------------------------------------------------- */
/* MUTEX_LOCK():                                                    */
/*   Locks a mutex lock.                                            */
/*                                                                  */
/* Argument:                                                        */
/*   MUTEX_t Id     -- the mutex ID to be locked                    */
/* Return value:                                                    */
/*   mtu_NORMAL     -- Success                                      */
/*   mtu_ERROR      -- Failure                                      */
/* ---------------------------------------------------------------- */

int MUTEX_LOCK(MUTEX_t MutexID)
{
     if (SYSTEM_INITIALIZE == FALSE) {
       fprintf(stderr, "Error in MUTEX_LOCK() ");
       fprintf(stderr, "MUTEX hasn't been initialized yet\n");
       exit(0);
     }

     if (MutexID->Magic != mtu_VALID) {        /* invalid mutex ID  */
          mtuMTP_errno = mtuMTP_INVALMUTEX;
          return mtu_ERROR;
     }
     if (MutexID->Owner != (void *)mtu_NOBODY) { /* locked?         */
          if (MutexID->Owner == Running) { /* lock recursively?     */
               mtuMTP_errno = mtuMTP_LOCKBYOWNER;
               return mtu_ERROR;
          }
          THREAD_Append(&(MutexID->WaitingQ), (void *)Running);
          Running->Status = mtu_WAITING;     /* queue the caller &  */
          THREAD_SCHEDULER ();               /* reschedule          */
     }
     MutexID->Owner = Running;          /* lock is open, lock it    */
     return mtu_NORMAL;
}

/* ---------------------------------------------------------------- */
/* MUTEX_UNLOCK():                                                  */
/*   Unlocks a mutex lock.                                          */
/*                                                                  */
/* Argument:                                                        */
/*   MUTEX_t Id     -- the mutex ID to be unlocked                  */
/* Return value:                                                    */
/*   mtu_NORMAL     -- Success                                      */
/*   mtu_ERROR      -- Failure                                      */
/* ---------------------------------------------------------------- */

int MUTEX_UNLOCK(MUTEX_t MutexID)
{
     THREAD_t thread;

     if (SYSTEM_INITIALIZE == FALSE) {
       fprintf(stderr, "Error in MUTEX_UNLOCK() ");
       fprintf(stderr, "MUTEX hasn't been initialized yet\n");
       exit(0);
     }

     if (MutexID->Magic != mtu_VALID) {        /* invalud mutex ID  */
          mtuMTP_errno = mtuMTP_INVALMUTEX;
          return mtu_ERROR;
     }
     if (MutexID->Owner == Running) {   /* owner try to unlock?     */
          MutexID->Owner = (THREAD_t) mtu_NOBODY; /* unlock it      */
          thread = (THREAD_t)THREAD_Remove (&(MutexID->WaitingQ));
          if (thread != NULL) {         /* let another thread in    */
               thread->Status = mtu_READY;
               MutexID->Owner = thread;
               THREAD_READY (thread);
               THREAD_YIELD ();
          }
          return mtu_NORMAL;
     }
     else {                             /* otherwise return error   */
          mtuMTP_errno = mtuMTP_NOTOWNER;
          return mtu_ERROR;
     }
}


