/* ---------------------------------------------------------------- */
/* THREAD_CREATE():                                                 */
/*   Creates a new thread.                                          */
/*                                                                  */
/* Arguments:                                                       */
/*   Entry          -- thread entry point (i.e., function to be run */
/*                     as a thread.                                 */
/*   Size           -- thread stack size                            */
/*   Flag           -- THREAD_SUSPENDED means that the new thread   */
/*                     suspended after it created.                  */
/*   Argc, Argv     -- thread function's parameters                 */
/* Return values:                                                   */
/*   mtu_ERROR      -- if failure                                   */
/*   New thread's ID -- if success                                  */
/* ---------------------------------------------------------------- */

THREAD_t THREAD_CREATE (void (*Entry)(), int Size, int Flag, 
            int Argc, char **Argv)
{
     THREAD_t NewThread;
     int      *StackBottom, FromAddr;
     void     *StackPointer;

     if (SYSTEM_INITIALIZE == FALSE)
         THREAD_SYS_INIT();

     NewThread = (THREAD_t )malloc (sizeof(TCB)); /* new thread TCB */
     if (NewThread == NULL) 
          return (THREAD_t) mtu_ERROR;
     Size += sizeof(StackAlign);        /* get new stack size       */
     StackBottom = (int *) malloc (Size);
     if (StackBottom == NULL) 
          return (THREAD_t) mtu_ERROR;  /* setup stack pointer      */
     StackPointer = (void *)(growsdown (&FromAddr)?
                (Size+(int)StackBottom)&-sizeof(StackAlign):(int)StackBottom);
     THREAD_INIT (NewThread, StackPointer);  /* initialize thread   */
                                        /* architecture-dependent!  */
     NewThread->Name   = NextThreadName++;/* initial TCB values     */
     NewThread->Status = mtu_READY;
     NewThread->Entry  = (void(*)(int, char**))Entry;
     NewThread->Argc   = Argc;
     NewThread->Argv   = Argv;
     NewThread->StackBottom = StackBottom;
     NewThread->Size   = Size;
     NewThread->JoinList.Head = NULL;
     NewThread->JoinList.Tail = NULL;

     THREAD_READY(NewThread);                /* thread into Ready Q */
     THREAD_READY(Running);                  /* add new and running */

     if (Flag == THREAD_SUSPENDED)
       THREAD_SUSPEND(NewThread);

     THREAD_SCHEDULER();                     /* reschedule threads  */

     return NewThread;
}

/* ---------------------------------------------------------------- */
/* THREAD_SELF():                                                   */
/*   Returns the ID of the running thread.                          */
/* ---------------------------------------------------------------- */

THREAD_t THREAD_SELF (void)
{
    if (SYSTEM_INITIALIZE == FALSE)
        THREAD_SYS_INIT();

    return Running;
}

/* ---------------------------------------------------------------- */
/* THREAD_YIELD():                                                  */
/*   Yields the processor to another thread.                        */
/* ---------------------------------------------------------------- */

void THREAD_YIELD (void)
{       
    if (SYSTEM_INITIALIZE == FALSE) {
        fprintf(stderr, "Error in THREAD_YIELD() ");
        fprintf(stderr, "No thread is running\n");
        exit(0);
    }

    THREAD_READY(Running);
    THREAD_SCHEDULER();
}

/* ---------------------------------------------------------------- */
/* THREAD_EXIT():                                                   */
/*   Terminates the current threads and frees its TCB.              */
/*                                                                  */
/* Return value:                                                    */
/*   mtu_ERROR      -- Failure                                      */
/*   mtu_NORMAL     -- current thread is the main thread and success*/
/*   no return value -- otherwise                                   */
/* ---------------------------------------------------------------- */

int THREAD_EXIT (void)
{
     THREAD_t temp;

     if (SYSTEM_INITIALIZE == FALSE) {
         fprintf(stderr, "Error in THREAD_EXIT() ");
         fprintf(stderr, "No thread is running\n");
         exit(0);
     }

     if (Running->Name == mtu_MAIN) {   /* if main, exit there      */
                                        /* have no thread remain    */
          if (ReadyQ.Head != NULL)      /* in the ready Q           */
               return mtu_ERROR;  
          if (SuspendQ.Head != NULL)    /* and in suspend queue     */
               return mtu_ERROR;
          return mtu_NORMAL;
     }

     while (Running->JoinList.Head != NULL) { /* check for joining  */
          temp = (THREAD_t)THREAD_Remove(&(Running->JoinList));
          temp->Status = mtu_READY;           /* make them ready    */
          THREAD_Append(&ReadyQ, (void *)temp);
     }

        
     Running->Name = mtu_INVALID;       /* set current thread's TCB */
     Running->Status = mtu_TERMINATED;  /* and status = terminated  */
     Running = NULL;
     THREAD_SCHEDULER();                /* run next thread          */
     return mtu_ERROR;                  /* shouldn't reach here     */
}

/* ---------------------------------------------------------------- */
/* THREAD_JOIN():                                                   */
/*   Causes the current process to wait for another thread's        */
/* termination.                                                     */
/*                                                                  */
/* Argument:                                                        */
/*   THREAD_t ThreadID   -- the thread being joined                 */
/* Return value:                                                    */
/*   mtu_ERROR           -- Failure                                 */
/*   mtu_NORMAL          -- Success                                 */
/* ---------------------------------------------------------------- */

int THREAD_JOIN (THREAD_t ThreadID)
{                  
          /* ------------------------------- */
          /* insert your implementation here */
          /* ------------------------------- */
     
     return mtu_NORMAL;
}

/* ---------------------------------------------------------------- */
/* THREAD_SUSPEND():                                                */
/*   Suspend the indicated thread.                                  */
/*                                                                  */
/* Argument:                                                        */
/*   THREAD_t ThreadID   -- the thread to be suspended              */
/* Return value:                                                    */
/*   mtu_ERROR           -- Failure                                 */
/*   mtu_NORMAL          -- Success                                 */
/* ---------------------------------------------------------------- */

int THREAD_SUSPEND (THREAD_t ThreadID)
{
          /* ------------------------------- */
          /* insert your implementation here */
          /* ------------------------------- */

     return mtu_NORMAL;
}

/* ---------------------------------------------------------------- */
/* THREAD_RESUME():                                                 */
/*   Resumes a suspended thread.                                    */
/*                                                                  */
/* Argument:                                                        */
/*   THREAD_t ThreadID   -- the thread to be resumed                */
/* Return value:                                                    */
/*   mtu_ERROR           -- Failure                                 */
/*   mtu_NORMAL          -- Success                                 */
/* ---------------------------------------------------------------- */

int THREAD_RESUME (THREAD_t ThreadID)
{
          /* ------------------------------- */
          /* insert your implementation here */
          /* ------------------------------- */

     return mtu_NORMAL;
}
