/* ---------------------------------------------------------------- */
/* PROGRAM  psort :                                                 */
/*    This program generates a Generator who keeps sending random   */
/* numbers to the first Sorter task, a Terminator who waits until   */
/* it received the END message and then exits, and many Sorter tasks*/
/* each of which has a mailbox connected with the previous Sorter   */
/* task, inChan, and a channel connected with its successor,outChan.*/
/* The Generator has a channel connected with the first Sorter's    */
/* inChan, while the Terminator has a channel connected with the
/* last Sorter's outChan.                                           */
/*    For each Sorter, it first gets a message from its inChan and  */
/* memorizes it.  Then, each Sorter keeps receiving messages from   */
/* its inChan.  If the incoming message if not an END, it is an     */
/* integer.  If this integer is larger than or equal to the number  */
/* that the Sorter memorized, the incoming number is sent to this   */
/* Sorter's neighbor via its outChan.  If this integer is smaller   */
/* than the number the Sorter's memorized, the new number is kept   */
/* and the memorized one is sent to the neighbor.                   */
/*    It is not difficult to see that when everyone receives the END*/
/* message, the numbers that are memorized by the Sorters are in    */
/* increasing order.  That is, they are sorted.                     */
/*                                                                  */
/*    This channel version of psort is ported directly from the     */
/* mbox version. Note that mbox is an AsynManyToManyChannel with    */
/* capacity of one.                                                 */
/* ---------------------------------------------------------------- */

#include  <stdio.h>
#include  <stdlib.h>
#include  <time.h>
#include  "mtuThread.h"

#define   END      -1                    /* the END message         */
#define   MAX_NUM   10                   /* max # sent by Generator */
#define   BOUND    100                   /* max value of random #s  */

/* ---------------------------------------------------------------- */
/*                    Function Prototypes                           */
/* ---------------------------------------------------------------- */

void  Generator(int);                    /* the Generator           */
void  Terminator(int);                   /* the Terminator          */
void  Sorter(int);                       /* the Sorters             */

/* ---------------------------------------------------------------- */
/*                        Channel                                   */
/* ---------------------------------------------------------------- */

CHANNEL_t Chan[MAX_NUM+1];

/* ---------------------------------------------------------------- */
/* FUNCTION  Generator:                                             */
/*    Function Generator sends random numbers to channel Chan[0].   */
/* After that it sends an END so that everybody would know to stop. */
/* ---------------------------------------------------------------- */

void  Generator(int ID)
{
    int i;
    int val;

    for (i = 0; i < MAX_NUM; i++) {       /* send numbers           */
        val = rand() % BOUND;
        CHAN_SEND(Chan[0], (void*) &val, sizeof(int));
        THREAD_YIELD();                 /* let other threads to run */
    }
    val = END;
    CHAN_SEND(Chan[0], (void*) &val, sizeof(int)); /* send END      */

    THREAD_EXIT();
}

/* ---------------------------------------------------------------- */
/* FUNCTION  Terminator:                                            */
/*    Function Terminator waits for the END message and then exits. */
/* Thus, stop all tasks.                                            */
/* ---------------------------------------------------------------- */

void  Terminator(int ID)
{
    int *msg;
    int size;

    CHAN_RECV(Chan[MAX_NUM], (void**) &msg, &size);

    if (*msg!=END)   /* waits msg     */
        printf("Oops, how could the Terminator get \
                a non-END (%d)?\n", *msg);
    printf("\n");
    exit(0);                            /* it must be END.  Exit    */
}

/* ---------------------------------------------------------------- */
/* FUNCTION  Sorter:                                                */
/*   Each Sorter has an inChan and an outChan. It receives the first*/
/* number and memorizes the number.  Then, each Sorter iterates and */
/* receives new numbers.  If the new one is not smaller than the    */
/* one this Sorter memorizes, the new one is sent via the outChan to*/
/* this Sorter's successor; otherwise, this Sorter memorizes the new*/
/* number and sends the memorized one.  At the end, this Sorter     */
/* displays the number it memorized and returns.                    */
/* ---------------------------------------------------------------- */

void  Sorter(int n)
{
    int inChan  = n - 1;                  /* my inChan              */
    int outChan = n;                      /* my outChan             */
    int val;
    int size;                             /* dummy var to recv size */
    int *MyNumber;                        /* the number I memorized */
    int *NewNumber;                       /* the incoming number    */

                                            /* receive my first no. */
                                            /* and get the next one */
    CHAN_RECV(Chan[inChan], (void**) &MyNumber, &size);    
    CHAN_RECV(Chan[inChan], (void**) &NewNumber, &size);   
    while (*NewNumber != END) {         /* while it is not the END  */
        if (*NewNumber >= *MyNumber)    /* if it is larger, then    */
                                                    /* send it      */
            CHAN_SEND(Chan[outChan], (void*)NewNumber, sizeof(int));   
        else {                                     /* otherwise,    */
                                                   /* send old      */
            CHAN_SEND(Chan[outChan], (void*)MyNumber, sizeof(int));    
            *MyNumber = *NewNumber;    /* and keep the new one      */
        }
                                                   /* receive next  */
        CHAN_RECV(Chan[inChan], (void**) &NewNumber, &size);    
    }
    printf("%4d", *MyNumber);           /* END received! display    */
    val = END;                          /* pass the END to my neigh.*/
    CHAN_SEND(Chan[outChan], (void*) &val, sizeof(int));

    THREAD_EXIT();
}

/* ---------------------------------------------------------------- */
/*                The main program starts here                      */
/* ---------------------------------------------------------------- */

int main(int argc, char *argv[])
{
    THREAD_t  id[MAX_NUM+2];              /* thread ID              */
    int      i, No[MAX_NUM+2];

    srand((unsigned int) time(NULL));   /* init. random # generator */

    for (i = 0; i < MAX_NUM+2; i++)       /* initialize task ID     */
        No[i] = i;

    for (i = 0; i <= MAX_NUM; i++)        /* create channel         */
    {
        char ch_name[20];
        sprintf(&ch_name[0], "Channel%d", i);
        Chan[i] = CHAN_INIT(ch_name, ASYNCHRONOUS, MANYTOMANY);
    }
                                          /* create the generator   */
    id[0] = THREAD_CREATE(Generator, THREAD_SIZE, THREAD_NORMAL, 
                No[0], (char **)0);

    for (i = 1; i < MAX_NUM+1; i++)       /* create sorters         */
        id[i] = THREAD_CREATE(Sorter, THREAD_SIZE, THREAD_NORMAL, 
                No[i], (char **)0);

    for (i = 1; i < MAX_NUM+1; i++)       /* join sorters           */
        THREAD_JOIN(id[i]);
                                          /* create the terminator  */
    id[MAX_NUM+1] = THREAD_CREATE(Terminator, THREAD_SIZE, THREAD_NORMAL, 
                No[MAX_NUM+1], (char **)0);
    THREAD_JOIN(id[MAX_NUM+1]);         /* never successful.  why?  */

    return 0;
}

